@@ -355,6 +355,11 @@ QQmlListProperty<QObject> QQuick3DViewport::data()
     This property specifies which \l Camera is used to render the scene. If this
     property is not set, then the first enabled camera in the scene will be used.
 
+    \note It is strongly recommended to explicitly set this property and not rely
+    on automatic camera selection. If there are multiple cameras in the scene, automatic
+    camera selection does not provide any guarantees regarding which camera will be selected.
+    If \l{Node::layer}{layers} are used, explicitly setting the camera is required.
+
     \note If this property contains a camera that's not \l {Node::visible}{visible} then
     no further attempts to find a camera will be done.
 
@@ -426,6 +431,14 @@ QQuick3DNode *QQuick3DViewport::scene() const
     referenced node becomes a sibling with child nodes of View3D, if there are
     any.
 
+    \note Scenes can only be shared between View3D items that are in the same
+    \l{QQuickWindow}{window}.
+
+    \note When sharing scenes between multiple View3D items the imported scene should
+    be imported in whole, that is, importing a subtree of a scene is not supported.
+    If multiple View3Ds need to show different parts of the same shared scene,
+    consider using \l {Node::layer}{layers} instead.
+
     \note This property can only be set once, and subsequent changes will have
     no effect.
 
@@ -566,8 +579,10 @@ QQuick3DSceneRenderer *QQuick3DViewport::createRenderer() const
             }
         }
 
-        if (rci)
+        if (rci) {
             renderer = new QQuick3DSceneRenderer(rci);
+            Q_QUICK3D_PROFILE_ASSIGN_ID(this, renderer);
+        }
     }
 
     return renderer;
@@ -1239,6 +1254,35 @@ QQuick3DPickResult QQuick3DViewport::rayPick(const QVector3D &origin, const QVec
     return getNearestPickResult(resultList);
 }
 
+/*!
+    \qmlmethod pickResult View3D::closestPointPick(vector3d origin, float radius, Model model)
+
+    This method will find the point on the surface of \a model that is nearest to \a origin, within a distance of
+    \a radius. If \a model is \c null, the closest object within \a radius will be found.
+
+    If no such object exists, \c null is returned.
+
+     \since 6.11
+*/
+QQuick3DPickResult QQuick3DViewport::closestPointPick(const QVector3D &origin, float radius, QQuick3DModel *model) const
+{
+    QQuick3DSceneRenderer *renderer = getRenderer();
+
+    if (!renderer)
+        return QQuick3DPickResult();
+    QSSGRenderNode *renderNode = nullptr;
+    if (model) {
+        renderNode = static_cast<QSSGRenderNode *>(QQuick3DObjectPrivate::get(model)->spatialNode);
+        if (Q_UNLIKELY(!renderNode))
+            return QQuick3DPickResult{};
+    }
+
+    const auto pickResult = renderer->syncPickClosestPoint(origin, radius * radius, renderNode);
+    if (!pickResult.has_value())
+        return QQuick3DPickResult{};
+    return processPickResult(pickResult.value());
+}
+
 void QQuick3DViewport::processPointerEventFromRay(const QVector3D &origin, const QVector3D &direction, QPointerEvent *event) const
 {
     internalPick(event, origin, direction);
@@ -1306,18 +1350,24 @@ void QQuick3DViewport::setTouchpoint(QQuickItem *target, const QPointF &position
 
     QPointingDevicePrivate *devPriv = QPointingDevicePrivate::get(m_syntheticTouchDevice);
 
-    auto makePoint = [devPriv](int id, QEventPoint::State pointState, QPointF pos) -> QEventPoint {
+    auto makePoint = [devPriv](int id, QEventPoint::State pointState, QPointF pos, quint64 timestamp) -> QEventPoint {
         auto epd = devPriv->pointById(id);
         auto &ep = epd->eventPoint;
         if (pointState != QEventPoint::State::Stationary)
             ep.setAccepted(false);
 
-        auto res = QMutableEventPoint::withTimeStamp(0, id, pointState, pos, pos, pos);
+        auto res = QMutableEventPoint::withTimeStamp(timestamp, id, pointState, pos, pos, pos);
         QMutableEventPoint::update(res, ep);
+
+        if (pointState == QEventPoint::State::Pressed)
+            QMutableEventPoint::setGlobalPressPosition(res, pos);
+        else if (ep.state() != QEventPoint::State::Unknown)
+            QMutableEventPoint::setGlobalPressPosition(res, ep.globalPressPosition());
+
         return res;
     };
 
-    auto sendTouchEvent = [&](QQuickItem *t, const QPointF &position, int pointId, QEventPoint::State pointState) -> void {
+    auto sendTouchEvent = [&](QQuickItem *t, const QPointF &position, int pointId, QEventPoint::State pointState, quint64 timestamp) -> void {
         QList<QEventPoint> points;
         bool otherPoint = false; // Does the event have another point already?
         for (int i = 0; i < m_touchState.size(); ++i) {
@@ -1325,11 +1375,11 @@ void QQuick3DViewport::setTouchpoint(QQuickItem *target, const QPointF &position
             if (ts.target != t)
                 continue;
             if (i == pointId) {
-                auto newPoint = makePoint(i, pointState, position);
+                auto newPoint = makePoint(i, pointState, position, timestamp);
                 points << newPoint;
             } else if (ts.isPressed) {
                 otherPoint = true;
-                points << makePoint(i, QEventPoint::Stationary, ts.position);
+                points << makePoint(i, QEventPoint::Stationary, ts.position, timestamp);
             }
         }
 
@@ -1342,6 +1392,7 @@ void QQuick3DViewport::setTouchpoint(QQuickItem *target, const QPointF &position
             type = QEvent::Type::TouchUpdate;
 
         QTouchEvent ev(type, m_syntheticTouchDevice, {}, points);
+        ev.setTimestamp(timestamp);
 
         if (t) {
             // Actually send event:
@@ -1361,13 +1412,15 @@ void QQuick3DViewport::setTouchpoint(QQuickItem *target, const QPointF &position
         }
     };
 
+    auto timestamp = QDateTime::currentMSecsSinceEpoch();
+
     // Send a release event to the previous target
     if (prevState.target && !sameTarget)
-        sendTouchEvent(prevState.target, prevState.position, pointId, QEventPoint::Released);
+        sendTouchEvent(prevState.target, prevState.position, pointId, QEventPoint::Released, timestamp);
 
     // Now send an event for the new state
     QEventPoint::State newState = isPress ? QEventPoint::Pressed : isRelease ? QEventPoint::Released : QEventPoint::Updated;
-    sendTouchEvent(target, position, pointId, newState);
+    sendTouchEvent(target, position, pointId, newState, timestamp);
 }
 
 QQuick3DLightmapBaker *QQuick3DViewport::maybeLightmapBaker()
@@ -1926,7 +1979,7 @@ bool QQuick3DViewport::singlePointPick(QSinglePointEvent *event, const QVector3D
     return delivered;
 }
 
-QPair<QQuickItem *, QPointF> QQuick3DViewport::getItemAndPosition(const QSSGRenderPickResult &pickResult)
+QPair<QQuickItem *, QPointF> QQuick3DViewport::getItemAndPosition(const QSSGRenderPickResult &pickResult) const
 {
     QQuickItem *subsceneRootItem = nullptr;
     QPointF subscenePosition;
@@ -2022,14 +2075,18 @@ QQuick3DPickResult QQuick3DViewport::processPickResult(const QSSGRenderPickResul
     QQuick3DObject *frontendObject = findFrontendNode(pickResult.m_hitObject);
 
     QQuick3DModel *model = qobject_cast<QQuick3DModel *>(frontendObject);
-    if (model)
+    if (model) {
+        auto itemAndPosition = getItemAndPosition(pickResult);
         return QQuick3DPickResult(model,
                                   ::sqrtf(pickResult.m_distanceSq),
                                   pickResult.m_localUVCoords,
                                   pickResult.m_scenePosition,
                                   pickResult.m_localPosition,
                                   pickResult.m_faceNormal,
-                                  pickResult.m_instanceIndex);
+                                  pickResult.m_sceneNormal,
+                                  pickResult.m_instanceIndex,
+                                  itemAndPosition.first);
+    }
 
     QQuick3DItem2D *frontend2DItem = qobject_cast<QQuick3DItem2D *>(frontendObject);
     if (frontend2DItem && frontend2DItem->contentItem()) {
@@ -2128,7 +2185,8 @@ void QQuick3DViewport::updateCameraForLayer(const QQuick3DViewport &view3D, QSSG
         for (QQuick3DCamera *camera : std::as_const(view3D.m_multiViewCameras))
             layerNode.explicitCameras.append(static_cast<QSSGRenderCamera *>(QQuick3DObjectPrivate::get(camera)->spatialNode));
     } else if (view3D.camera()) {
-        layerNode.explicitCameras.append(static_cast<QSSGRenderCamera *>(QQuick3DObjectPrivate::get(view3D.camera())->spatialNode));
+        if (QSSGRenderCamera *camera = static_cast<QSSGRenderCamera *>(QQuick3DObjectPrivate::get(view3D.camera())->spatialNode))
+            layerNode.explicitCameras.append(camera);
     }
 
     // Ensure these have a parent. All nodes need to be in the node tree somewhere, even if they're technically "parentless"
@@ -2160,7 +2218,7 @@ void QQuick3DViewport::updateSceneManagerForImportScene()
 
     }
     connect(privateObject->sceneManager, &QQuick3DSceneManager::needsUpdate,
-            this, &QQuickItem::update);
+            this, &QQuickItem::update, Qt::UniqueConnection);
     connect(privateObject->sceneManager, &QObject::destroyed,
             this, [&](QObject *) {
                 auto privateObject = QQuick3DObjectPrivate::get(m_importScene);
@@ -2176,7 +2234,7 @@ void QQuick3DViewport::updateSceneManagerForImportScene()
         if (scene) {
             connect(QQuick3DObjectPrivate::get(scene)->sceneManager,
                     &QQuick3DSceneManager::needsUpdate,
-                    this, &QQuickItem::update);
+                    this, &QQuickItem::update, Qt::UniqueConnection);
         }
     }
 }
